/* Name: usb162.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2007-12-9
 * Tabsize: 8
 * Copyright: (c) 2007-2009 by Koji Suzuki (suz@nmj2.sakura.ne.jp)
 * License: Proprietary, free under certain conditions. See Documentation.
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <usbconfig.h>
#include "usb162.h"
#include "desc.h"

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#define nop()	__asm__ volatile ("nop" )


#ifndef USB162_NO_APPEP
#define USB162_NO_APPEP		0
#endif

#if defined(__AVR_ATmega16U4__) || defined(__AVR_ATmega32U4__)
#define MAX_EP		6
#else
#define MAX_EP		4
#endif

#define TYPE_CONT	0
#define TYPE_BULK	2
#define TYPE_INT	3

#define SIZE_8		0

#define ONE_BANK	0
#define DUAL_BANK	1

#define EP_CTRL		0

static uint8_t ep_ctrl_size;
int8_t usb_connected;
#ifndef USB162_NO_USER_SETUP
uint8_t (* usb_user_setup)(uint8_t *, uint8_t *);
#endif
#ifndef USB162_NO_VENDOR_SETUP
uint8_t (* usb_vendor_setup)(uint8_t *, uint8_t *);
#endif
#ifndef USB162_NO_USER_OUT
uint8_t (* usb_user_out)(uint8_t *, uint8_t);
#endif

static inline void send_zlp () {
	bit_clear(UEINTX, TXINI);  // send ZLP 
	while (bit_is_clear(UEINTX, TXINI))
		;
}

static int endpoint_setup(uint8_t ep) {
	char *p;
	uint8_t size,sz;
#if !USB162_NO_APPEP
	uint8_t dir,attr,limit,bank;
	if (ep == EP_CTRL) {
#endif
		p = usbDescrDevice;
		size = __LPM(p+7);
		sz = SIZE_8;
		while ((8<<sz) < size )
			sz++;
		UENUM = EP_CTRL;
		if (bit_is_clear(UECONX, EPEN)) {
			bit_set(UECONX, EPEN);
			UECFG0X = (TYPE_CONT<<EPTYPE0) | (0<<EPDIR);
			UECFG1X = (sz<<EPSIZE0) | (ONE_BANK<<EPBK0);
			bit_set(UECFG1X,ALLOC);
		} else {
			bit_set(UERST,EP_CTRL);
			nop();
			bit_clear(UERST,EP_CTRL);
		}
		ep_ctrl_size = size;
		return size;
#if !USB162_NO_APPEP
	}
	p = usbDescrConfig;
	limit = __LPM(p+2);
	while (p - usbDescrConfig < limit) {
		if ((__LPM(p+1) == USB_DESCR_ENDPOINT) 
		     && ((__LPM(p+2) & 0xf) == ep))  { // endpoint desc.
			dir = (__LPM(p+2) & 0x80) == 0x80; // 0x80 : IN
			attr = __LPM(p+3); // bulk:2 , interrupt:3
			size = __LPM(p+4);
			sz = SIZE_8;
			while ((8<<sz) < size )
				sz++;
			bank = ONE_BANK;
			if ((ep >= 3) && (attr == 2)) {
#if (USE_DUAL_BANK_BULK_IN)
				if (dir) {
					bank = DUAL_BANK;
				}
#endif
#if (USE_DUAL_BANK_BULK_OUT)
				if (!dir) {
					bank = DUAL_BANK;
				}
#endif
			}
			UENUM = ep;
			if (bit_is_clear(UECONX, EPEN)) {
				bit_set(UECONX, EPEN);
				UECFG0X = (attr<<EPTYPE0) | dir;
				UECFG1X = (sz<<EPSIZE0) | (bank<<EPBK0);
				bit_set(UECFG1X,ALLOC);
			} else {
				bit_set(UERST,ep);
				nop();
				bit_clear(UERST,ep);
			}
			return size;
		}
		p += __LPM(p);
	}
#endif
	return 0;
}

#ifndef F_SCALER
#define F_SCALER 0
#endif

#ifndef PLLP0
#define PLLP0	PINDIV
#endif

static void start_pll() {
#if defined(__AVR_ATmega16U4__) || defined(__AVR_ATmega32U4__)
		// 96 MHz
		//PLLFRQ = ((1 << PLLUSB) | (1 << PDIV3) | (1 << PDIV1));
#endif
	if (bit_is_clear(PLLCSR,PLOCK)) {
#if (F_CPU << F_SCALER) == 16000000UL
		PLLCSR = (1<<PLLP0) | (1<<PLLE);
#else
		PLLCSR = (0<<PLLP0) | (1<<PLLE);
#endif
		while (bit_is_clear(PLLCSR,PLOCK))
			;
	}
	bit_clear(USBCON,FRZCLK); // enable usb clock 
}

#ifdef EP_MS_OUT
extern void callback_msc_reset();
#endif

static inline void usbReset() {
#if USB162_SOFT_RESET
#if !USB162_NO_APPEP
	int8_t i;
#endif

	usb_connected = 0;
	GPIOR2 = 0;
	bit_set(USBCON,USBE);     // enable usb
#if !USB162_NO_APPEP
	for (i=0; i<= MAX_EP; i++) {
		endpoint_setup(i);
	}
#else
	endpoint_setup(0);
#endif
#ifdef EP_MS_OUT
	callback_msc_reset(1);
#endif
#endif
}

void usbInit() {
#if !USB162_NO_APPEP
	int8_t i;
#endif
	USBCON = (1<<FRZCLK);
	UDIEN = 0;
	UEIENX = 0;
	usb_connected = 0;
	GPIOR2 = 0;
#if defined(__AVR_ATmega16U4__) || defined(__AVR_ATmega32U4__)
	bit_set(UHWCON,UVREGE);     // usb pad power enable
	//_delay_ms(1);
	bit_set(USBCON,OTGPADE);     // enable usb pad
#endif
#if 0 //defined(__AVR_ATmega16U4__) || defined(__AVR_ATmega32U4__)
	while (bit_is_clear(USBSTA, VBUS))
		wdt_reset();
#endif
	bit_set(USBCON,USBE);     // enable usb
	bit_clear(UDCON, DETACH); // connect internal pull-up 
	start_pll();
	while (bit_is_clear(UDINT, SOFI))
		wdt_reset();
	UDINT = 0;	// for reentrance from USB reset int (EORSTI)
#if USB162_SOFT_RESET
	bit_clear(UDCON, RSTCPU);
#else
	bit_set(UDCON, RSTCPU);   // reset cpu @ USB_RESET
#endif
	//bit_clear(UDADDR, ADDEN); // meaning-less
#if !USB162_NO_APPEP
	for (i=0; i<= MAX_EP; i++) {
		endpoint_setup(i);
	}
#else
	endpoint_setup(0);
#endif
	//bit_set(UDIEN, SUSPE);    // Enable Suspend int
	bit_set(UDIEN, EORSTE);	  // Enable Reset int
	//bit_set(UDIEN, WAKEUPE);  // Enable Wakeup int (one time)
}

#if USB162_SOFT_RESET
/* __bad_interrupt  means the following:
extern void __init();
ISR(USB_GEN_vect, ISR_NAKED) {
	__init();
}
*/
#else
ISR(USB_GEN_vect) {
	if (bit_is_set(UDINT,EORSTI)) {
		// USB Reset
		usbReset();
		bit_set(GPIOR2, USB162_RECV_RESET);
		bit_clear(UDINT,EORSTI);
	}
	if (bit_is_set(UDINT, WAKEUPI)) {
		bit_clear(UDIEN, WAKEUPE);
		start_pll();
		bit_clear(UDINT, WAKEUPI);
	}
}
#endif

static void usb_reply_setup(uint8_t requested, uint8_t *p, uint8_t len, int8_t mode) {
	uint8_t fifo_len = 0;
	uint8_t written = 0;

	bit_clear(UEINTX, NAKOUTI);
	while (len) {
		if (bit_is_set(UEINTX, NAKOUTI)) { // got ZLP -- abort
			goto ret;
		}
		UEDATX = (mode == 0)?(*p):(__LPM(p));
		p++;
		fifo_len++;
		written++;
		len--;
		if (written >= requested) break;
		if (fifo_len == ep_ctrl_size) {
			bit_clear(UEINTX, TXINI); // send
			while (bit_is_clear(UEINTX, TXINI)
					&& bit_is_clear(UEINTX, NAKOUTI))
				;
			fifo_len = 0;
		}
	}
	if (bit_is_set(UEINTX, NAKOUTI)) { // got ZLP -- abort
		goto ret;
	}
ret:
	if (fifo_len || (written != requested)) { // send ZLP
		bit_clear(UEINTX, TXINI);
	}
	// wait ZLP (from HOST)
	while (bit_is_clear(UEINTX, RXOUTI)
		&& bit_is_clear(UEINTX, NAKOUTI))
		;
	bit_clear(UEINTX, NAKOUTI);
	bit_clear(UEINTX, RXOUTI);
}

#ifndef MAX_REPLY_LEN
#define MAX_REPLY_LEN 8
#endif
static uint8_t reply_data[MAX_REPLY_LEN];
static uint8_t setup_data[8];

static uint8_t usb_config_value;

static inline void usb_process_setup() {
	uint8_t reply_requested;
	uint8_t request;
	uint8_t request_type;
	uint8_t rcpt;
	uint8_t r;
	uint8_t *p;

	bit_clear(UEINTX,RXOUTI);
	for (r=0; r<8; r++) {
		if (!UEBCLX) {
			bit_set(UECONX, STALLRQ);
			return;
		}
		setup_data[r] = UEDATX;
	}
	request_type = setup_data[USB_RQ_TYPE];
	request = setup_data[USB_RQ_REQ];
	bit_clear(UEINTX,RXSTPI);

	reply_requested = setup_data[USB_RQ_LENGTH_L];
	if (setup_data[USB_RQ_LENGTH_H]) {
		reply_requested = 254;
	}
	if ((request_type & USB_RQ_TYPE_MASK) == USB_RQ_TYPE_STANDARD) {
		if (request == USB_RQ_GET_STATUS) {
			rcpt = request_type & USB_RQ_RCPT_MASK;
			reply_data[0] = 0;
			reply_data[1] = 0;
			if (rcpt == USB_RQ_RCPT_DEVICE) {
				reply_data[0] = (__LPM(usbDescrConfig+7) & USB_ATTR_SELFPOWER)?1:0;
			}
			if (rcpt == USB_RQ_RCPT_ENDPOINT 
				&& setup_data[USB_RQ_INDEX_L] == 0x81)
				reply_data[0] = 1;
			usb_reply_setup(reply_requested, reply_data, 2, 0);
		} else if (request == USB_RQ_SET_ADDRESS) {
#if 0
			if (!setup_data[USB_RQ_VALUE_L] && !UDADDR) {
				bit_set(UECONX, STALLRQ);
				return;
			}
#endif
			UDADDR = setup_data[USB_RQ_VALUE_L] & 0x7f;
			send_zlp();

			bit_set(UDADDR, ADDEN); // enable addrss
		} else if (request == USB_RQ_GET_DESCRIPTOR) {
			if ( setup_data[USB_RQ_VALUE_H] == 1) {
				usb_reply_setup(reply_requested,
						(uint8_t *)usbDescrDevice,
						__LPM(usbDescrDevice), 1);
			} else if (setup_data[USB_RQ_VALUE_H] == 2) {
				usb_reply_setup(reply_requested,
						(uint8_t *)usbDescrConfig,
						__LPM(usbDescrConfig+2), 1);
			} else if (setup_data[USB_RQ_VALUE_H] == 3) {
#if 1
			    p = (void *)__LPM_word(usbDescrStringTable + 
			    		(setup_data[USB_RQ_VALUE_L] & 3));
			    usb_reply_setup(reply_requested, (uint8_t *)p,
						__LPM(p), 1);
#else
			    if (setup_data[USB_RQ_VALUE_L] == 0) {
				usb_reply_setup(reply_requested,
						(uint8_t *)usbDescrString0,
						__LPM(usbDescrString0), 1);
			    } else if (setup_data[USB_RQ_VALUE_L] == 1) {
				usb_reply_setup(reply_requested,
						(uint8_t *)usbDescrString1,
						__LPM(usbDescrString1), 1);
			    } else if (setup_data[USB_RQ_VALUE_L] == 2) {
				usb_reply_setup(reply_requested,
						(uint8_t *)usbDescrString2,
						__LPM(usbDescrString2), 1);
			    } else if (setup_data[USB_RQ_VALUE_L] == 3) {
				usb_reply_setup(reply_requested,
						(uint8_t *)usbDescrString3,
						__LPM(usbDescrString3), 1);
			    }
#endif
#if USB_CFG_DESCR_PROPS_HID
			} else if (setup_data[USB_RQ_VALUE_H] == USB_DESCR_HID) {
				usb_reply_setup(reply_requested,
						(uint8_t *)usbDescrConfig + 18,
						9, 1);
#endif
			} else {
				bit_set(UECONX, STALLRQ);
			}
		} else if (request == USB_RQ_GET_CONFIGURATION) {
			reply_data[0] = usb_config_value;
			usb_reply_setup(reply_requested, reply_data, 1, 0);
		} else if (request == USB_RQ_SET_CONFIGURATION) {
			usb_config_value = setup_data[USB_RQ_VALUE_L];
			bit_clear(UEINTX, TXINI);  // send ZLP ?
			usb_connected = 1;
		} else if (request == USB_RQ_GET_INTERFACE) {
			reply_data[0] = 0;
			usb_reply_setup(reply_requested, reply_data, 1, 0);
		} else if (request == USB_RQ_CLEAR_FEATURE || request == USB_RQ_SET_FEATURE) {
			if (setup_data[USB_RQ_VALUE_L] == 0 &&
				setup_data[USB_RQ_INDEX_L] == 0x81) {
				/* EP1 HALT */
				bit_set(UECONX, STALLRQ);
			}
			bit_clear(UEINTX, TXINI); // send ZLP ?
		} else if (request == USB_RQ_SET_INTERFACE) {
			bit_clear(UEINTX, TXINI); // send ZLP ?
		} else {
			bit_set(UECONX, STALLRQ);
		}
	} else {
		r = 254;
#ifndef USB162_NO_VENDOR_SETUP
		if (((request_type & USB_RQ_TYPE_MASK) == USB_RQ_TYPE_VENDOR) 
			&& usb_vendor_setup) {
			r = (usb_vendor_setup)(setup_data,reply_data);
#else
		if (0) {
#endif
#ifndef USB162_NO_USER_SETUP
		} else if (usb_user_setup) {
			r = (usb_user_setup)(setup_data,reply_data);
#endif
		}
		if (r == 255) {
			send_zlp();
		} else if (r == 254) {
			bit_set(UECONX, STALLRQ);
		} else if ((request_type & USB_RQ_DIR_MASK) == USB_RQ_DIR_DEVICE_TO_HOST) {
			usb_reply_setup(reply_requested, reply_data, r, 0);
		}
	}
}

#ifndef USB162_NO_USER_OUT
static inline void usb_process_out() {
	uint8_t r;

	for (r=0; r<8; r++) {
		if (!UEBCLX) break;
		setup_data[r] = UEDATX;
	}
	bit_clear(UEINTX,RXOUTI);
	if (usb_user_out) {
	    r = (usb_user_out)(setup_data,r);
	} else {
	    r = 254;
	}
	if (r == 255) send_zlp();
	else if (r == 254) bit_set(UECONX, STALLRQ);
}
#endif

void usbPoll() {
	if (bit_is_set(GPIOR2, USB162_RECV_RESET)) {
		bit_clear(GPIOR2, USB162_RECV_RESET);
	}
	bit_clear(UDINT, SOFI);
	UENUM = EP_CTRL;
	if (bit_is_set(UEINTX, RXSTPI)) { // recieve SETUP 
		usb_process_setup();
	}
#ifndef USB162_NO_USER_OUT
	else if (bit_is_set(UEINTX, RXOUTI)) {
		usb_process_out();
	}
#endif
	if (bit_is_set(UEINTX,NAKINI)) {
		bit_clear(UEINTX,NAKINI);
	}
}
