/*
 *	Copyright (c)2010	Koji Suzuki
 */

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <isp.h>
#include <pdi.h>
#include <tpi.h>
#include <clock.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

#define TX_ERROR		(-1)
#define RX_NODATA		(-1)
#define RX_ERROR_PARITY		0x100
#define RX_ERROR_FRAMING	0x200
#define RX_ERROR_OVERRUN	0x400

#include <aspconfig.h>

#define RECV_TIMEOUT		2	// 4-8ms
#define SEND_TIMEOUT		20 //2	// 4-8ms

#define NVMBUSY_TIMEOUT		5	// 16-20ms
#define CHIPERASE_TIMEOUT	10	// 36-40ms

#define SEND_BUFSIZ		32
#define RECV_BUFSIZ		32

#if defined(ENABLE_TPI) || defined(ENABLE_PDI)

static int8_t is_tpi;
static uchar pditpi_enabled;
static uchar pditpi_need_break;
static uchar pditpi_sck;
static uchar pditpi_delay_c; // (>= USBASP_ISP_SCK_32 @ 16MHz)
			     // in SOFT_XXX , it means 1/2 xck cpu_clocks
			     // in SOFT_PDITPI it means nothing (fixed 16k bps)
			     // in HARD_PDITPI , it means 16xck timer_clocks

static volatile uint8_t	timeout1_timer;
static volatile uint8_t	timeout2_timer;

static const uint8_t key_val[8] = { 0xff,0x88,0xd8,0xcd,0x45,0xab,0x89,0x12 };

static void pditpi_wait_xck(void);
static void pditpi_oflush(void);
static void pditpi_idle(void); // end of tx/rx
static void pditpiSetSCKOption(uchar option);
static void pditpiHWdisable(void);
static void pditpiHWenable(void);
static int8_t pditpi_send(uint8_t data);
static int16_t pditpi_recv(void);

#ifdef ENABLE_SOFT_XXX
  #define USBASP_ISP_SCK_MAX     USBASP_ISP_SCK_500
  #define USBASP_ISP_SCK_MIN     USBASP_ISP_SCK_16
  #define USBASP_ISP_SCK_DEFAULT USBASP_ISP_SCK_16
#elif defined(ENABLE_SOFT_PDITPI)
  #define USBASP_ISP_SCK_MAX     USBASP_ISP_SCK_16
  #define USBASP_ISP_SCK_MIN     USBASP_ISP_SCK_16
  #define USBASP_ISP_SCK_DEFAULT USBASP_ISP_SCK_16
#else
  #define USBASP_ISP_SCK_MAX     USBASP_ISP_SCK_8000
  #define USBASP_ISP_SCK_MIN     USBASP_ISP_SCK_32
  #define USBASP_ISP_SCK_DEFAULT USBASP_ISP_SCK_250
#endif

static void pditpiSetSCKOption(uchar option) {
	if ((option == USBASP_ISP_SCK_AUTO) || ( option > USBASP_ISP_SCK_MAX)) 
		option = USBASP_ISP_SCK_DEFAULT;
	if (option < USBASP_ISP_SCK_MIN)
		option = USBASP_ISP_SCK_MIN;
#if F_CPU == 8000000UL
	if ( option > USBASP_ISP_SCK_4000) 
		option = USBASP_ISP_SCK_4000;
	UBRR1 = (1 << (14 - option)) - 1;
#elif F_CPU == 16000000UL
	UBRR1 = (1 << (15 - option)) - 1;
#elif F_CPU == 12000000UL
	if ( option > USBASP_ISP_SCK_2000) 
		option = USBASP_ISP_SCK_2000;
	UBRR1 = (3 << (13 - option)) - 1;
#endif
	pditpi_sck = option;
#ifdef ENABLE_SOFT_XXX
  #if F_CPU == 8000000UL
	pditpi_delay_c = (2 << (USBASP_ISP_SCK_8000 - option - 4)) - 1;
  #elif F_CPU == 12000000UL
	pditpi_delay_c = (3 << (USBASP_ISP_SCK_8000 - option - 4)) - 1;
  #elif F_CPU == 16000000UL
	pditpi_delay_c = (4 << (USBASP_ISP_SCK_8000 - option - 4)) - 1;
  #endif
#else
  #if F_CPU == 8000000UL
	pditpi_delay_c = (2 << (USBASP_ISP_SCK_8000 - option - 3)) + 1;
  #elif F_CPU == 12000000UL
	pditpi_delay_c = (3 << (USBASP_ISP_SCK_8000 - option - 3)) + 1;
  #elif F_CPU == 16000000UL
	pditpi_delay_c = (4 << (USBASP_ISP_SCK_8000 - option - 3)) + 1;
  #endif
#endif
}

#ifdef ENABLE_SOFT_XXX

static volatile uint8_t soft_xck_enable;
static uint8_t timer_count;

static uchar pditpi_delay_c2;

/* 4 clock version */
static inline void __delay_clk(uint8_t __count) {
        __asm__ volatile (
                "1: dec %0" "\n\t"
                "nop" "\n\t"
                "brne 1b"
                : "=r" (__count)
                : "0" (__count)
        );
}

static uint8_t pditpi_clk() {
	uint8_t r = 0;
    	bit_clear(PDITPI_OUT,PDITPI_SCK);
	__delay_clk(pditpi_delay_c);
    	bit_set(PDITPI_OUT,PDITPI_SCK);
	if (bit_is_set(PDITPI_IN,PDITPI_MOSI))
		r |= 0x80;
	__delay_clk(pditpi_delay_c2);
	return r;
}

ISR(TIMER1_COMPA_vect) {
	uint8_t state = soft_xck_enable;
	timer_count++;
	if (state == 1) { // starting
 		if (timer_count & 1) {
    			bit_clear(PDITPI_OUT,PDITPI_SCK);
			soft_xck_enable = 2;
		}
	} else if (state == 2) {
 		if (timer_count & 1) {
    			bit_clear(PDITPI_OUT,PDITPI_SCK);
		} else {
    			bit_set(PDITPI_OUT,PDITPI_SCK);
		}
	} else if (state == 3) { // stopping
 		if (!(timer_count & 1)) {
    			bit_set(PDITPI_OUT,PDITPI_SCK);
		} else {
			soft_xck_enable = 0;
		}
	}
	if ((timer_count & 0x7f) == 0) { // 32KHz / 128 = 250Hz
	    if (timeout1_timer) {
		--timeout1_timer;
	    }
	    if (timeout2_timer) {
		--timeout2_timer;
	    }
	}
}

static void pditpi_start_xck() {
	while (soft_xck_enable & 1)
		;
	if (!soft_xck_enable) {
		soft_xck_enable = 1;
	}
}

static void pditpi_stop_xck() {
	while (soft_xck_enable & 1)
		;
	if (soft_xck_enable) {
		soft_xck_enable = 3;
		while (soft_xck_enable)
			;
	}
}

static void pditpi_wait_xck() {
	uint8_t count = 15;

#if 1 // for test only
	pditpi_start_xck();
	while (--count) {
		while (bit_is_set(PDITPI_OUT, PDITPI_SCK))
				;
		while (bit_is_clear(PDITPI_OUT, PDITPI_SCK))
				;
	}
#else
	pditpi_stop_xck();
	while (--count) {
		pditpi_clk();
	}
#endif
}

static void pditpi_iflush() {
	if (!pditpi_dir_is_tx()) {
		bit_set(PDITPI_OUT,PDITPI_MOSI);
		bit_set(PDITPI_DDR,PDITPI_MOSI);
		pditpi_dir_tx();
		if (pditpi_need_break) {
			bit_clear(PDITPI_OUT,PDITPI_MOSI);
			pditpi_wait_xck();
			bit_set(PDITPI_OUT,PDITPI_MOSI);
			pditpi_wait_xck();
			pditpi_need_break = 0;
		}
	}
	pditpi_start_xck();
}

static void pditpi_idle() {
	if (pditpi_enabled) {
		pditpi_iflush();
	}
}

static void pditpiHWdisable() {
	if (pditpi_enabled)
		pditpi_stop_xck();

	/* timer stop */
	bit_clear(TIMSK1,OCIE1A);
	TCCR1B = 0;
	OCR1A = 0;

	if (is_tpi) {
		tpi_reset(0);
	}

    	bit_set(PDITPI_OUT,PDITPI_MOSI);
	bit_clear(PDITPI_DDR,PDITPI_MOSI);  // TX = pull-up 
    	bit_set(PDITPI_OUT,PDITPI_SCK);
    	bit_clear(PDITPI_DDR,PDITPI_SCK);  // XCK = pull-up
	pditpi_enabled = 0;
}

static void pditpiHWenable() {
	pditpi_delay_c2 = pditpi_delay_c - 6;
	if ((int8_t)pditpi_delay_c2 < 0) pditpi_delay_c2 = 0;
	pditpi_dir_init();

	if (is_tpi) {
		tpi_reset(1);
		clockWait(200); // 64ms
		clockWait(200); // 64ms
	}
#ifdef PDITPI_MISO
	bit_clear(PDITPI_DDR,PDITPI_MISO);
	bit_clear(PDITPI_OUT,PDITPI_MISO);  // RX = HI-Z
#endif
	bit_set(PDITPI_OUT,PDITPI_MOSI);
	bit_set(PDITPI_DDR,PDITPI_MOSI); // TX = H
    	bit_set(PDITPI_OUT,PDITPI_SCK);
    	bit_set(PDITPI_DDR,PDITPI_SCK);  // XCK = H

	clockWait(1); // 320us wait before enable XCK

	if (!pditpi_enabled)
		soft_xck_enable = 0;
	// 32 kHz timer = 16kHz XCK
	TCCR1C = 0;
	OCR1A =  (F_CPU/1000/32 - 1) & 0xffff;
	TCCR1A = 0;
	TCCR1B = (1<<WGM12)|(1<<CS10);  // CTC  1/1
	bit_set(TIMSK1,OCIE1A);

	pditpi_wait_xck();
	pditpi_wait_xck();
	pditpi_start_xck();
	pditpi_enabled = 1;
}


static int8_t pditpi_send(uint8_t data) {
	uint8_t i;
	uint8_t parity = 0;

	pditpi_iflush();
	pditpi_stop_xck();
	bit_clear(PDITPI_OUT,PDITPI_MOSI); // start bit
	pditpi_clk();
	for (i=0; i<8; i++) {
		if (data & 1) {
			bit_set(PDITPI_OUT,PDITPI_MOSI);
		} else {
			bit_clear(PDITPI_OUT,PDITPI_MOSI);
		}
		parity ^= data;
		pditpi_clk();
		data >>= 1;
	}
	if (parity & 1) {
		bit_set(PDITPI_OUT,PDITPI_MOSI);
	} else {
		bit_clear(PDITPI_OUT,PDITPI_MOSI);
	}
	pditpi_clk();
	bit_set(PDITPI_OUT,PDITPI_MOSI); // stop bit
	pditpi_clk();
	pditpi_clk();
	return 0;
}

static int16_t pditpi_recv() {
	uint8_t i;
	uint8_t data = 0;
	uint8_t parity = 0;
	int16_t r;

	pditpi_stop_xck();
	if (pditpi_dir_is_tx()) {
		bit_clear(PDITPI_DDR,PDITPI_MOSI);
		pditpi_dir_rx();
		bit_clear(PDITPI_OUT,PDITPI_MOSI); // TX = HI-Z
	}
	timeout1_timer = RECV_TIMEOUT;
//retry:
	while (timeout1_timer) {
		data = pditpi_clk();
		if (data == 0) {
			break;
		}
	}
	if (data & 0x80) {
		pditpi_need_break |= (RX_NODATA >> 8);
		return RX_NODATA; // TIMEOUT
	}
	for (i=0; i<8; i++) {
		data >>= 1;
		data |= pditpi_clk();
		parity ^= data;
	}
	r = data;
	parity ^= pditpi_clk();
	if (parity & 0x80) {
		r |= RX_ERROR_PARITY;
	}
	parity = 0;
	if (!(pditpi_clk() & 0x80)) {
		r |= RX_ERROR_OVERRUN;
		parity |= 0x80;
	}
	if (!(pditpi_clk() & 0x80)) {
		r |= RX_ERROR_OVERRUN;
		parity |= 0x80;
	}
	if ((data == 0) && (parity & 0x80)) { // break;
	    r |= RX_ERROR_FRAMING;
	    while (timeout1_timer) {
		if (pditpi_clk()) {
			break;
		}
	    }
	}
#if 0
	if (r & RX_ERROR_PARITY) {
		if (!is_tpi && (data == 0xbb || data == 0xdb)) {
			parity = 0;
			data = 0;
			goto retry;
		}
	}
#endif
	pditpi_need_break |= (r >> 8);
	return r;
}

#elif defined(ENABLE_SOFT_PDITPI)

static volatile uint8_t soft_xck_enable;

static volatile uint16_t send_data;
static volatile int16_t recv_buf[RECV_BUFSIZ];
static volatile uint8_t	recv_buf_in = 0;
static volatile uint8_t	recv_buf_out = 0;

static volatile int8_t rx_state;
static volatile int8_t tx_state;

static uint8_t timer_count;

ISR(TIMER1_COMPA_vect) {
	int8_t state;
	uint8_t v = 0;
	uint8_t next;
	uint16_t *p;

	timer_count++;
	if (soft_xck_enable) {
	    if (timer_count & 1) {
    		bit_clear(PDITPI_OUT,PDITPI_SCK);
		if (pditpi_dir_is_tx()) {
			state = tx_state;
			if (state) {
			    if (send_data & 1) {
				bit_set(PDITPI_OUT, PDITPI_MOSI);
			    } else {
				bit_clear(PDITPI_OUT, PDITPI_MOSI);
			    }
			    send_data >>= 1;
			    tx_state = state - 1;
			}
		}
	    } else {
    		bit_set(PDITPI_OUT,PDITPI_SCK);
		if (!pditpi_dir_is_tx()) {
		    if (bit_is_set(PDITPI_IN, PDITPI_MOSI)) {
			v |= 1;
		    }
 		    p = (uint16_t *)(&recv_buf[recv_buf_in]);
		    state = rx_state;
		    if (state == 0) { // start
			if (!(v & 1)) {
				*p = 0;
				rx_state = 1;
			}
		    } else if (state < 0) { // break
			if (v & 1) {
				rx_state = 0;
			}
		    } else if (state == 11) { // stop2
			if (*p == 0) {
				rx_state = (-1);
			} else {
				rx_state = 0;
			}
			next = (recv_buf_in + 1) & (RECV_BUFSIZ -1);
			if (next != recv_buf_out) {
				recv_buf_in = next;
			}
		    } else {
			*p >>= 1;
			if (v & 1) {
				*p |= 0x0200;
			}
			rx_state = state + 1;
		    }
		}
	    }
	}
	if ((timer_count & 0x7f) == 0) { // 32KHz / 128 = 250Hz
	    if (timeout1_timer) {
		--timeout1_timer;
	    }
	    if (timeout2_timer) {
		--timeout2_timer;
	    }
	}
}

static void pditpi_start_xck() {
	if (!soft_xck_enable) {
		soft_xck_enable = 1;
		while (bit_is_clear(PDITPI_IN, PDITPI_SCK))
			;
		while (bit_is_set(PDITPI_IN, PDITPI_SCK))
			;
	}
}

static void pditpi_stop_xck() {
	if (soft_xck_enable) {
		while (bit_is_clear(PDITPI_IN, PDITPI_SCK))
			;
		while (bit_is_set(PDITPI_IN, PDITPI_SCK))
			;
		cli();
		soft_xck_enable = 0;
		rx_state = tx_state = 0;
		recv_buf_in = recv_buf_out = 0;
		sei();
	}
}

static void pditpi_wait_xck() {
#if 0
	uint8_t count = 15;

	pditpi_start_xck();
	while (--count) {
		while (bit_is_clear(PDITPI_IN, PDITPI_SCK))
				;
		while (bit_is_set(PDITPI_IN, PDITPI_SCK))
				;
	}
#else
	pditpi_start_xck();
	// wait > (15 * 1000/16) us
	clockWait(4);
#endif
}

static void pditpi_iflush() {
	if (!pditpi_dir_is_tx()) {
		bit_set(PDITPI_OUT,PDITPI_MOSI);
		bit_set(PDITPI_DDR,PDITPI_MOSI);
		pditpi_dir_tx();
		if (pditpi_need_break) {
			bit_clear(PDITPI_OUT,PDITPI_MOSI);
			pditpi_wait_xck();
			bit_set(PDITPI_OUT,PDITPI_MOSI);
			pditpi_wait_xck();
			pditpi_need_break = 0;
		}
	}
}

static void pditpi_oflush() {
	wdt_reset();
	if (soft_xck_enable && pditpi_dir_is_tx()) {
		timeout1_timer = SEND_TIMEOUT;
		while (timeout1_timer) {
			if (tx_state == 0)
				break;
		}
		while (bit_is_clear(PDITPI_IN, PDITPI_SCK))
				;
		while (bit_is_set(PDITPI_IN, PDITPI_SCK))
				;
	}
}

static void pditpi_idle() {
	if (pditpi_enabled) {
		pditpi_iflush();
	}
}

static void pditpiHWdisable() {
	pditpi_stop_xck();

	/* timer stop */
	bit_clear(TIMSK1,OCIE1A);
	TCCR1B = 0;
	OCR1A = 0;

	if (is_tpi) {
		tpi_reset(0);
	}

    	bit_set(PDITPI_OUT,PDITPI_MOSI);
	bit_clear(PDITPI_DDR,PDITPI_MOSI);  // TX = pull-up 
    	bit_set(PDITPI_OUT,PDITPI_SCK);
    	bit_clear(PDITPI_DDR,PDITPI_SCK);  // XCK = pull-up
	pditpi_enabled = 0;
}

static void pditpiHWenable() {
	pditpi_dir_init();

	if (is_tpi) {
		tpi_reset(1);
		clockWait(200); // 64ms
		clockWait(200); // 64ms
	}
#ifdef PDITPI_MISO
	bit_clear(PDITPI_DDR,PDITPI_MISO);
	bit_clear(PDITPI_OUT,PDITPI_MISO);  // RX = HI-Z
#endif
	bit_set(PDITPI_OUT,PDITPI_MOSI);
	bit_set(PDITPI_DDR,PDITPI_MOSI); // TX = H
    	bit_set(PDITPI_OUT,PDITPI_SCK);
    	bit_set(PDITPI_DDR,PDITPI_SCK);  // XCK = H

	clockWait(1); // 320us wait before enable XCK

	// 32 kHz timer = 16kHz XCK
	TCCR1A = 0;
	TCCR1B = (1<<WGM12)|(1<<CS10);  // CTC  1/1
	TCCR1C = 0;
	OCR1A =  (F_CPU/1000/32 - 1) & 0xffff;
	bit_set(TIMSK1,OCIE1A);

	pditpi_wait_xck();
	pditpi_wait_xck();
	pditpi_enabled = 1;
}

/*
 * pditpi_send: data format
 *  MSB                                     LSB
 *   +---+---+---+--+--+--+--+--+--+--+--+---+
 *   | 1 | 1 | P |D7|D6|D5|D4|D3|D2|D1|D0| 0 |
 *   +---+---+---+--+--+--+--+--+--+--+--+---+
 *   | --------------- 13 bit ---------------|
 *
 * send_data = data;
 * tx_state = 12;
 */
static int8_t pditpi_send(uint8_t data) {
	uint8_t i;
	uint8_t parity = 0;

	pditpi_iflush();
	timeout1_timer = SEND_TIMEOUT;
	while (timeout1_timer && tx_state)
		;
	if (tx_state) return TX_ERROR;

	send_data = 0;

	for (i=0; i<8; i++) {
		send_data <<= 1;
		if (data & 0x80) {
			send_data |= 2; 
		}
		parity ^= data;
		data <<= 1;
	}
	if (parity & 0x80) {
		send_data |= 0x200;
	}
	send_data |= 0x400;
	send_data |= 0x800;
	tx_state = 12;
	return 0;
}

/*
 * pditpi_recv: data format
 *  MSB                                 LSB
 *   +---+---+--+--+--+--+--+--+--+--+
 *   | 1 | P |D7|D6|D5|D4|D3|D2|D1|D0|
 *   +---+---+--+--+--+--+--+--+--+--+
 *   | ----------- 10 bit -----------|
 *
 */
static int16_t pditpi_recv() {
	uint8_t ptr;
	uint8_t parity = 0;
	int16_t c = RX_NODATA;
	int16_t r;
	int8_t i;

	wdt_reset();
	if (pditpi_dir_is_tx()) {
		pditpi_oflush();
		rx_state = 0;
		recv_buf_in = recv_buf_out = 0;
		bit_clear(PDITPI_DDR,PDITPI_MOSI);
		pditpi_dir_rx();
		bit_clear(PDITPI_OUT,PDITPI_MOSI); // TX = HI-Z
	}
	timeout1_timer = RECV_TIMEOUT;
	while (timeout1_timer) {
	    ptr = recv_buf_out;
	    if (recv_buf_in != ptr) {
		c = recv_buf[ptr];
		ptr = (ptr + 1) & (SEND_BUFSIZ -1);
		recv_buf_out = ptr;
		break;
	    }
	    wdt_reset();
	}
	if (c < 0) {
		pditpi_need_break |= (RX_NODATA >> 8);
		return c;
	}
	if (c == 0) {
		pditpi_need_break |= (RX_ERROR_FRAMING >> 8);
		return RX_ERROR_FRAMING;
	}
	r = c & 0xff;
 	if (!(c & 0x200)) {
		r |= RX_ERROR_OVERRUN;
	}
	for (i=0; i<9; i++) {
		parity ^= c;
		c >>= 1;
	}
	if (parity & 1) {
		r |= RX_ERROR_PARITY;
	}
	pditpi_need_break |= (r >> 8);
	return (r);
}

#elif defined(ENABLE_HARD_PDITPI)

static volatile int16_t recv_buf[RECV_BUFSIZ];
static volatile uint8_t	recv_buf_in = 0;
static volatile uint8_t	recv_buf_out = 0;

static volatile uint8_t	timeout1_timer;
static volatile uint8_t	timeout2_timer;


static void pditpi_wait_xck() {
#if 0
	if (pditpi_sck < USBASP_ISP_SCK_1000) {
		uint8_t count = 16;
		while (--count) {
			while (bit_is_clear(PDITPI_IN, PDITPI_SCK))
					;
			while (bit_is_set(PDITPI_IN, PDITPI_SCK))
					;
		}
	} else {
		clockDelay(pditpi_delay_c);
	}
#else
	clockDelay(pditpi_delay_c);
#endif
}

static void pditpi_oflush() {
	wdt_reset();
	if (bit_is_set(UCSR1B,TXEN1)) {
		timeout1_timer = SEND_TIMEOUT;
		while (timeout1_timer) {
			if (bit_is_set(UCSR1A,TXC1)) {
				bit_set(UCSR1A,TXC1);
				break;
			}
		}
	}
}

static void pditpi_iflush() {
	if (bit_is_clear(UCSR1B,TXEN1)) {
		bit_set(UCSR1A, TXC1);
		bit_set(PDITPI_OUT,PDITPI_MOSI);
		bit_set(PDITPI_DDR,PDITPI_MOSI);
		pditpi_dir_tx();
		if (pditpi_need_break) {
			bit_clear(PDITPI_OUT,PDITPI_MOSI);
			pditpi_wait_xck();
			bit_set(PDITPI_OUT,PDITPI_MOSI);
			pditpi_wait_xck();
			pditpi_need_break = 0;
		}
		UCSR1B = (1<<TXEN1);
		cli();
		recv_buf_in = recv_buf_out = 0;
		sei();
	}
}
 
static void pditpi_idle() {
	if (pditpi_enabled) {
		pditpi_iflush();
	}
}

static void pditpiHWdisable() {
	pditpi_oflush();

	/* timer stop */
	bit_clear(TIMSK1,OCIE1A);
	TCCR1B = 0;
	OCR1A = 0;

	UCSR1B = 0;	      //disable interrputs
	UCSR1C = 0;

	if (is_tpi) {
		tpi_reset(0);
	}

    	bit_set(PDITPI_OUT,PDITPI_MOSI);
	bit_clear(PDITPI_DDR,PDITPI_MOSI);  // TX = pull-up 
    	bit_set(PDITPI_OUT,PDITPI_SCK);
    	bit_clear(PDITPI_DDR,PDITPI_SCK);  // XCK = pull-up

	// clear buffers
	cli();
	recv_buf_in = recv_buf_out = 0;
	sei();
	pditpi_enabled = 0;
}

static void pditpiHWenable() {
	pditpi_dir_init();

	if (is_tpi) {
		tpi_reset(1);
		clockWait(200); // 64ms
		clockWait(200); // 64ms
	}
	bit_clear(PDITPI_DDR,PDITPI_MISO);
	bit_clear(PDITPI_OUT,PDITPI_MISO);  // RX = HI-Z
	bit_set(PDITPI_OUT,PDITPI_MOSI);
	bit_set(PDITPI_DDR,PDITPI_MOSI); // TX = H
    	bit_set(PDITPI_OUT,PDITPI_SCK);
    	bit_set(PDITPI_DDR,PDITPI_SCK);  // XCK = H

	clockWait(1); // 320us wait before enable XCK

	bit_clear(UCSR1A , U2X1); // must clear in syncronus mode
	UCSR1C = (1<<UMSEL10) | (2<<UPM10) | (1<<USBS1)
			| (3<<UCSZ10) | (1<<UCPOL1);

	//enable interrputs
	UCSR1B = (1<<RXEN1)|(1<<RXCIE1);

	// 250Hz timer
	TCCR1A = 0;
	TCCR1B = (1<<WGM12)|(3<<CS10);  // CTC  1/64
	TCCR1C = 0;
	OCR1A =  (F_CPU/64/250 - 1) & 0xffff;
	bit_set(TIMSK1,OCIE1A);

	pditpi_wait_xck();
	pditpi_wait_xck();
	pditpi_enabled = 1;
}


ISR(TIMER1_COMPA_vect) {
	if (timeout1_timer) {
		--timeout1_timer;
	}
	if (timeout2_timer) {
		--timeout2_timer;
	}
}

ISR(USART1_RX_vect) {
	int16_t c;
	uint8_t flags,ptr,next;
	flags = UCSR1A;
	c = UDR1;
	if (bit_is_set(flags, UPE1)) {
#if 0
		if (!is_tpi && (c == 0xDB || c == 0xEB)) // DELAY or EMPTY
			return;
#endif
		c |= RX_ERROR_PARITY; // BREAK or real error
	}
	if (bit_is_set(flags, FE1)) {
		c |= RX_ERROR_FRAMING; // real break ?
	}
	if (bit_is_set(flags, DOR1)) {
		c |= RX_ERROR_OVERRUN;
	}
	ptr = recv_buf_in;
	next = (ptr + 1) & (RECV_BUFSIZ -1);
	if (next != recv_buf_out) {
		recv_buf[ptr] = c;
		recv_buf_in = next;
	}
}

static int8_t pditpi_send(uint8_t data) {
	pditpi_iflush();
	timeout1_timer = SEND_TIMEOUT;
	while (timeout1_timer) {
		if (bit_is_set(UCSR1A, UDRE1)) {
			UDR1 = data;
			return 0;
		}
		wdt_reset();
	}
	return TX_ERROR;
}

static int16_t pditpi_recv() {
	uint8_t ptr;
	int16_t c = RX_NODATA;

	wdt_reset();
	if (bit_is_set(UCSR1B,TXEN1)) {
		pditpi_oflush();
		cli();
		recv_buf_in = recv_buf_out = 0;
		sei();

		UCSR1B = (1<<RXEN1)|(1<<RXCIE1);
		bit_clear(PDITPI_DDR,PDITPI_MOSI);
		pditpi_dir_rx();
		bit_clear(PDITPI_OUT,PDITPI_MOSI); // TX = HI-Z
	}
	timeout1_timer = RECV_TIMEOUT;
	while (timeout1_timer) {
	    ptr = recv_buf_out;
	    if (recv_buf_in != ptr) {
		c = recv_buf[ptr];
		ptr = (ptr + 1) & (SEND_BUFSIZ -1);
		recv_buf_out = ptr;
		return (c);
	    }
	    wdt_reset();
	}
	pditpi_need_break |= (c >> 8);
#if SERIAL_DEBUG
if (pditpi_need_break) {
	debug_puts("ERROR pditpi_recv :");
	debug_put_hex(pditpi_need_break);
	debug_puts("\r\n");
}
#endif
	return (c);
}

#endif // ENABLE_SOFT_XXX

static void pditpiDisconnect() {
	/* disable hardware */
	pditpiHWdisable();
}

#ifdef ENABLE_TPI
/* TPI operations */

static uint16_t tpi_addr;

static int16_t tpiReadCS(unsigned int address) {
	uint8_t inst;
	int16_t r;
	inst = address & 0xf;
	if (address & 0x8000) { /* read CS */
		inst |= 0x80; // TPI_SLDCS
	} else { /* read IO */
		inst |= (address << 1) & 0x60;	
		inst |= 0x10; // TPI_SIN
	}

	r = pditpi_send(inst);
	if (!r) r = pditpi_recv();
	return r;
}

static uchar tpiWriteCS(unsigned int address, uchar data) {
	uint8_t inst;
	inst = address & 0xf;
	if (address & 0x8000) { /* write CS */
		inst |= 0xc0; // TPI_SSTCS
	} else {
		inst |= (address << 1) & 0x60;
		inst |= 0x90; // TPI_SOUT
	}
	pditpi_send(inst);
	pditpi_send(data);
	return 0;
}

static uchar tpiSetPR(uint16_t addr) {
	uint8_t inst;
	inst = 0x68; // TPI_SETPR_LO
	pditpi_send(inst);
	pditpi_send(addr);

	inst = 0x69; // TPI_SETPR_HI
	pditpi_send(inst);
	pditpi_send(addr >> 8);
	return 0;
	
}

static int16_t tpiWaitNVM() {
	int16_t r = -1;

	pditpi_oflush();
	timeout2_timer = NVMBUSY_TIMEOUT;
	while (timeout2_timer) {
		r = tpiReadCS(TPI_NVMCSR);
		if (!(r & 0xff00) && !((uchar)r & (1<< TPI_NVMBSY))) {
			return 0;
		}
		wdt_reset();
	}
	return r;
}

static int16_t tpiReadDS(unsigned long address) {
	uint8_t inst;
	int16_t r;
	if (address != tpi_addr) {
		tpiWriteCS(TPI_NVMCMD, TPI_NVM_NOP);
		tpiSetPR(address);
		tpi_addr = address;
	}
	inst = 0x24; // TPI_SLD_POSTI
	r = pditpi_send(inst);
	if (!r) r = pditpi_recv();
	tpi_addr ++;
	return r;
}

static uchar tpiWriteDS(unsigned long address, uchar data, uchar pollmode) {
	uint8_t inst;
	if (address != tpi_addr) {
		tpiSetPR(address);
		tpi_addr = address;
	}

	inst = 0x64; // TPI_SST_POSTI
	pditpi_send(inst);
	pditpi_send(data);
	if ((uint8_t)tpi_addr & 0x01) 
		tpiWaitNVM();
	tpi_addr ++;
	return 0;
}

static uchar tpiEnable() {
	int8_t i;
	uint8_t inst;
	const uint8_t *p;
	int16_t r;
	uchar count = 4;

	while (count--) {
		tpi_addr = 0xffff;

#if USBASP_ISP_SCK_MAX > USBASP_ISP_SCK_2000
		tpiWriteCS(TPI_CS_TPIPCR, TPI_CS_IDLE_16); // idle bits  = 2 + 8
		tpiReadCS(TPI_CS_TPIPCR);
		tpiWriteCS(TPI_CS_TPIPCR, TPI_CS_IDLE_16); // idle bits  = 2 + 8
		tpiReadCS(TPI_CS_TPIPCR);
#else
		tpiWriteCS(TPI_CS_TPIPCR, TPI_CS_IDLE_8); // idle bits  = 2 + 8
		tpiReadCS(TPI_CS_TPIPCR);
		tpiWriteCS(TPI_CS_TPIPCR, TPI_CS_IDLE_8); // idle bits  = 2 + 8
		tpiReadCS(TPI_CS_TPIPCR);
#endif

		inst = 0xe0;	// TPI_SKEY
		pditpi_send(inst);
		p = key_val;
		for (i=0; i < 8; i++) {
			pditpi_send(*p++);
		}
		i = 10;
		while (--i) {
			r = tpiReadCS(TPI_CS_TPISR);
			if (!(r & 0xff00) && ((uchar)r & (1<< TPI_CS_NVMEN))) {
				pditpi_idle();
				return 0;
			}
			wdt_reset();
		}
		pditpiHWdisable();
		pditpiHWenable();
	}
	return 1; /* error: device dosn't answer */
}

/* read and write 4 bytes from pditpi */
static void tpiTransmit4(uchar *out, uchar *in) {
	uint8_t tmp;
	uint8_t ispcmd = in[0];
	int16_t r = 0;
	uint16_t addr;

	addr = ((uint16_t)in[1]<<8) + in[2];

	tpi_addr = 0xffff;
        if(ispcmd == ISPCMD_READ_SIGNATURE) {
            tmp = in[2] & 3;
            r = tpiReadDS(TPI_DS_SIGNATURE + tmp);
        } else if(ispcmd == ISPCMD_WRITE_EEPROM) {
	    r = tpiWriteCS(addr,in[3]);
	    pditpi_oflush();
#if 1 // for digi-ana test
	    if (r >=0 ) r = tpiReadCS(addr);
#endif 
        } else if(ispcmd == ISPCMD_READ_EEPROM) {
	    r = tpiReadCS(addr);
        } else if(ispcmd == ISPCMD_CHK_BUSY) {
	    r = tpiWaitNVM();
        } else if(ispcmd == ISPCMD_ISP_CONF && in[2] == 0) {
	    if(in[1] == ISPCONF_CHIP_ERASE) {
		tpiWriteCS(TPI_NVMCMD, TPI_NVM_CHIP_ERASE);
		tpiWriteDS(TPI_DS_FLASH + 1, 0, 1); // HI-byte required
	        pditpi_oflush();
	    }
	}
	pditpi_idle();
        out[2] = (r >> 8);
        out[3] = r;
}
#endif

#endif
#ifdef ENABLE_PDI
/* PDI operations */

static uint32_t pdi_addr;
static int16_t pdi_pgmsk;
static uint8_t pdi_len;
static uint8_t blk_len;
static uint8_t pdi_space;

struct pdi_space_conf {
   uint8_t read;
   uint8_t clear_buff;
   uint8_t load_buff;
   uint8_t erase_page;
   uint8_t write_page;
};

static const struct pdi_space_conf pdi_space_conf[] = {
  /* space0: ALL_FLASH */
  { PDI_NVM_READ_NVM
  , PDI_NVM_CLEAR_FLASH_BUFF	//x
  , PDI_NVM_FILL_FLASH_BUFF	//x
  , PDI_NVM_ERASE_FLASH_PAGE	//none
  , PDI_NVM_WRITE_FLASH_PAGE } ,  //x

  /* space1: APP_FLASH */
  { PDI_NVM_READ_NVM
  , PDI_NVM_CLEAR_FLASH_BUFF	//x
  , PDI_NVM_FILL_FLASH_BUFF	//x
  , PDI_NVM_ERASE_APP_PAGE	//x
  , PDI_NVM_WRITE_APP_PAGE } , //x

  /* space2: BOOT_FLASH */
  { PDI_NVM_READ_NVM
  , PDI_NVM_CLEAR_FLASH_BUFF	//x
  , PDI_NVM_FILL_FLASH_BUFF	//x
  , PDI_NVM_ERASE_BOOT_PAGE	//x
  , PDI_NVM_WRITE_BOOT_PAGE } , //x

  /* space3: EEPROM */
  { PDI_NVM_READ_NVM //PDI_NVM_READ_EEPROM
  , PDI_NVM_CLEAR_EEPROM_BUFF	//x
  , PDI_NVM_FILL_EEPROM_BUFF	//x
  , PDI_NVM_ERASE_EEPROM_PAGE	//x
  , PDI_NVM_ERASEWRITE_EEPROM_PAGE //, PDI_NVM_WRITE_EEPROM_PAGE 
} , 

  /* space4: FUSE */
  { PDI_NVM_READ_NVM
  , PDI_NVM_NOP
  , PDI_NVM_WRITE_FUSE	//x
  , PDI_NVM_NOP
  , PDI_NVM_NOP } , 

  /* space5: LOCKBITS */
  { PDI_NVM_READ_NVM
  , PDI_NVM_NOP
  , PDI_NVM_WRITE_LOCKBIT //x
  , PDI_NVM_NOP
  , PDI_NVM_NOP } , 

  /* space6: USERSIG */
  { PDI_NVM_READ_NVM // PDI_NVM_READ_USERSIG
  , PDI_NVM_CLEAR_FLASH_BUFF
  , PDI_NVM_FILL_FLASH_BUFF
  , PDI_NVM_ERASE_USERSIG
  , PDI_NVM_WRITE_USERSIG } , //x

  /* space7: CALIBRATION */
  { PDI_NVM_READ_NVM //PDI_NVM_READ_CALIBRATION
  , PDI_NVM_NOP
  , PDI_NVM_NOP
  , PDI_NVM_NOP
  , PDI_NVM_NOP } , 

};

#define PDI_NVM_READ		(pdi_space_conf[pdi_space].read)
#define PDI_NVM_CLEAR_BUFF	(pdi_space_conf[pdi_space].clear_buff)
#define PDI_NVM_FILL_BUFF	(pdi_space_conf[pdi_space].load_buff)
#define PDI_NVM_ERASE_PAGE	(pdi_space_conf[pdi_space].erase_page)
#define PDI_NVM_WRITE_PAGE	(pdi_space_conf[pdi_space].write_page)

/* memo : 
 PDI_NVM_READ_NVM can read from all (?)
*/

static inline int16_t pdiReadNVM(uint16_t addr) {
	uint8_t inst;
	int16_t r;
	inst = PDI_LDS41;

	r = pditpi_send(inst);
	if (!r) r = pditpi_send(addr);
	if (!r) r = pditpi_send(addr>>8);
	if (!r) r = pditpi_send(0x00);
	if (!r) r = pditpi_send(0x01);
	if (!r) r = pditpi_recv();
	return r;
}

static inline uchar pdiWriteNVM(uint16_t addr, uchar data) {
	uint8_t inst;
	inst = PDI_STS41;
	pditpi_send(inst);
	pditpi_send(addr);
	pditpi_send(addr>>8);
	pditpi_send(0x00);
	pditpi_send(0x01);
	pditpi_send(data);
	return 0;
}

static int16_t pdiReadCS(uint16_t address) {
	uint8_t inst;
	int16_t r;
	if (!(address & 0x8000)) return pdiReadNVM(address);
	inst = address & 0xf;
	inst |= PDI_LDCS;

	r = pditpi_send(inst);
	if (!r) r = pditpi_recv();
	return r;
}

static uchar pdiWriteCS(uint16_t address, uchar data) {
	uint8_t inst;
	if (!(address & 0x8000)) return pdiWriteNVM(address,data);
	inst = address & 0xf;
	inst |= PDI_STCS;
	pditpi_send(inst);
	pditpi_send(data);
	return 0;
}

static uchar pdiSetPR(uint32_t addr) {
	uint8_t inst;
	inst = PDI_SETPR4;
	pditpi_send(inst);
	pditpi_send(addr);
	pditpi_send(addr >> 8);
	pditpi_send(addr >> 16);
	pditpi_send(addr >> 24);
	return 0;
}

static int16_t pdiWaitNVM() {
	int16_t r = -1;

	pditpi_oflush();
	timeout2_timer = NVMBUSY_TIMEOUT;
	while (timeout2_timer) {
		r = pdiReadCS(PDI_NVM_STATUS);
		if (!(r & 0xff00) &&
		    !((uchar)r & ((1<< PDI_NVM_BUSY)|(1<< PDI_NVM_FBUSY)))) {
			return 0;
		}
		wdt_reset();
	}
	return r; // error
}


static void pdiStartBlk(uchar len, int16_t pgmsk) {
	pdi_len = len;
	pdi_pgmsk = pgmsk;
	blk_len = 0;
}

static int16_t pdiReadDS(unsigned long address) {
	uint8_t inst;
	int16_t r;
	if (!blk_len && (address != pdi_addr)) {
		pdiWriteCS(PDI_NVM_CMD, PDI_NVM_READ);
		pdiSetPR(address);
		pdi_addr = address;
	}
	if (!blk_len && pdi_len) {
		blk_len = pdi_len;
		if (blk_len > RECV_BUFSIZ) blk_len = RECV_BUFSIZ;
		pdi_len -= blk_len;
		if (blk_len) {
			inst = PDI_REPEAT1;
			pditpi_send(inst);
			pditpi_send(blk_len -1);

			inst = PDI_LD1_POSTI;
			pditpi_send(inst);
		}
	}
#ifdef SERIAL_DEBUG
	debug_puts("pdiReadDS(");
	debug_put_hex(pdi_addr >> 24);
	debug_put_hex(pdi_addr >> 16);
	debug_put_hex(pdi_addr >> 8);
	debug_put_hex(pdi_addr);
	debug_puts(") , b ");
	debug_put_hex(blk_len);
	debug_puts("\r\n");
#endif
	if (blk_len) {
		blk_len--;
	} else {
		inst = 0x24; // PDI_LD1_POSTI
		pditpi_send(inst);
	}
	r = pditpi_recv();
	pdi_addr ++;
	return r;
}

static uchar pdiWriteDS(unsigned long address, uchar data, uchar pollmode) {
	uint8_t inst;
	if (!blk_len && (address != pdi_addr)) {
		pdi_addr = address;
		pdiSetPR(pdi_addr);
	}
	if (!blk_len && pdi_len) {
		blk_len = pdi_len;
		if (blk_len > SEND_BUFSIZ) blk_len = SEND_BUFSIZ;
		if (pdi_pgmsk > 0) {
			uint16_t tmp = (uint16_t)pdi_addr & pdi_pgmsk;
			tmp = pdi_pgmsk + 1 - tmp;
			if (blk_len > tmp) blk_len = tmp;
		}
		pdi_len -= blk_len;
		if (blk_len) {
			inst = 0xa0; // PDI_REPEAT1
			pditpi_send(inst);
			pditpi_send(blk_len -1);

			inst = 0x64; // PDI_ST1_POSTI
			pditpi_send(inst);
		}
	}
	if (blk_len) {
		blk_len--;
	} else {
		inst = 0x64; // PDI_ST1_POSTI
		pditpi_send(inst);
	}
	pditpi_send(data);
	pdi_addr ++;
	return 0;
}

static uchar pdiEnable() {
	int8_t i;
	uint8_t inst;
	const uint8_t *p;
	int16_t r;
	uchar count = 4;

	while (count--) {
		pdi_len = 0;
		blk_len = 0;
		pdi_addr = 0xffffffff;


#if USBASP_ISP_SCK_MAX > USBASP_ISP_SCK_2000
		pdiWriteCS(PDI_CS_CTRL, PDI_CS_IDLE_16); // idle bits  = 2 + 8
		pdiReadCS(PDI_CS_CTRL);
		pdiWriteCS(PDI_CS_CTRL, PDI_CS_IDLE_16); // idle bits  = 2 + 8
		pdiReadCS(PDI_CS_CTRL);
#else
		pdiWriteCS(PDI_CS_CTRL, PDI_CS_IDLE_8); // idle bits  = 2 + 8
		pdiReadCS(PDI_CS_CTRL);
		pdiWriteCS(PDI_CS_CTRL, PDI_CS_IDLE_8); // idle bits  = 2 + 8
		pdiReadCS(PDI_CS_CTRL);
#endif

		pdiWriteCS(PDI_CS_RESET, PDI_CS_RESET_KEY);
		i = 10;
		while (--i && (pdiReadCS(PDI_CS_RESET) != 0x00)) {
			wdt_reset();
		}
		inst = 0xe0;	// PDI_KEY
		pditpi_send(inst);
		p = key_val;
		for (i=0; i < 8; i++) {
			pditpi_send(*p++);
		}
		i = 10;
		while (--i) {
			r = pdiReadCS(PDI_CS_STATUS);
			if (r == (1<< PDI_CS_NVMEN)) {
				pditpi_idle();
				return 0;
			}
			wdt_reset();
		}
		pditpiHWdisable();
		pditpiHWenable();
	}
	return 1; /* error: device dosn't answer */
}

static uchar pdiOpenPage(unsigned long address, uchar pollvalue) {
	uint8_t sav_len;
	sav_len = pdi_len;
	pdi_len = 0;
#if 0
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_ERASE_PAGE);
	pdiWriteDS(address , 0, 0);
	pdiWaitNVM();
#endif
#if 0
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_CLEAR_BUFF);
	pdiWriteCS(PDI_NVM_CTRLA, (1<<PDI_NVM_CMDEX));
	pdiWaitNVM();
#endif
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_FILL_BUFF);
	pdi_len = sav_len;
	return 0;
}

static uchar pdiClosePage(unsigned long address, uchar pollvalue) {
	uint8_t sav_len = pdi_len;
	uint16_t tmp = (uint16_t)address & pdi_pgmsk;
	pdi_len = 0;

	if ((pdi_pgmsk > 0) && tmp) {
		tmp = pdi_pgmsk + 1 - tmp;
		pdi_len = tmp;
		while(--tmp) {
			++address;
			pdiWriteDS(address , 0xff, 0);
		}
	}
#if 0
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_ERASE_PAGE);
	pdiWriteDS(address , 0, 0);
	pdiWaitNVM();
#endif
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_WRITE_PAGE);
	pdiWriteDS(address , 0, 0);
	pdiWaitNVM();
	pdiWriteCS(PDI_NVM_CMD, PDI_NVM_NOP);
	pdi_len = sav_len;
	return 0;
}

/* read and write 4 bytes from pditpi */
static void pdiTransmit4(uchar *out, uchar *in) {
	int16_t r = 0;
	uint16_t addr;
	int8_t tmp;
	uint8_t ispcmd = in[0];
	addr = ((uint16_t)in[1]<<8) + in[2];

	pdi_len = 0;
	blk_len = 0;
	pdi_addr = 0xffffffff;

        if (ispcmd == ISPCMD_READ_SIGNATURE) {
            tmp = in[2] & 3;
            r = pdiReadCS(PDI_MCU_DEVID0 + tmp);
        } else if(ispcmd == ISPCMD_WRITE_EEPROM) {
	    r = pdiWriteCS(addr,in[3]);
	    pditpi_oflush();
#if SERIAL_DEBUG
	debug_puts("pdiWriteCS(");
	debug_put_hex((addr >> 8)&0xff);
	debug_put_hex(addr & 0xff);
	debug_puts(",");
	debug_put_hex(in[3]);
	debug_puts(") = ");
	
	debug_put_hex((r >> 8)&0xff);
	debug_put_hex(r & 0xff);
	debug_puts("\r\n");
#endif
        } else if(ispcmd == ISPCMD_READ_EEPROM) {
	    r = pdiReadCS(addr);
#ifdef SERIAL_DEBUG
	debug_puts("pdiReadCS(");
	debug_put_hex((addr >> 8)&0xff);
	debug_put_hex(addr & 0xff);
	debug_puts(") = ");
	
	debug_put_hex((r >> 8)&0xff);
	debug_put_hex(r & 0xff);
	debug_puts("\r\n");
#endif
        } else if(ispcmd == ISPCMD_CHK_BUSY) {
	    r = pdiWaitNVM();
#if SERIAL_DEBUG
	debug_puts("pdiWaitNVM = ");
	debug_put_hex((r >> 8)&0xff);
	debug_put_hex(r & 0xff);
	debug_puts("\r\n");
#endif
        } else if(ispcmd == ISPCMD_ISP_CONF && in[2] == 0) {
	    if(in[1] == ISPCONF_CHIP_ERASE) {
		pdiWriteCS(PDI_NVM_CMD, PDI_NVM_CHIP_ERASE);
		pdiWriteCS(PDI_NVM_CTRLA, (1<<PDI_NVM_CMDEX));
	        pditpi_oflush();
	    } else if(in[1] == ISPCONF_SET_SPACE) {
		pdi_space = in[3] & 0x7;
#if SERIAL_DEBUG
	debug_puts("pdi_space = ");
	debug_put_hex(pdi_space);
	debug_puts("\r\n");
#endif
	    } else if(in[1] == ISPCONF_RESET) {
		pdiWriteCS(PDI_CS_RESET, in[3]) ;
		timeout2_timer = NVMBUSY_TIMEOUT;
		while (timeout2_timer && (pdiReadCS(PDI_CS_RESET) != 0x00)) {
			wdt_reset();
		}
	    }
        } else {
            /* ignore all others, return default value == 0 */
        }
	pditpi_idle();
        out[2] = (r >> 8);
        out[3] = r;
}
#endif

void pditpiConnect(struct pgm_ops *ops, uchar sck_opt, uchar __is_tpi) {
#if defined(ENABLE_TPI) || defined(ENABLE_PDI)
	pditpiSetSCKOption(sck_opt);
	is_tpi = __is_tpi;

	pditpiHWenable();

	ops->pgm_disconnect   = pditpiDisconnect;
	if (is_tpi) {
#ifdef ENABLE_TPI
		ops->pgm_poll         = pditpi_idle;
		ops->pgm_transmit     = tpiTransmit4;
		ops->pgm_enable       = tpiEnable;
		ops->pgm_read_space0  = (uchar (*)(uint16_t))tpiReadCS;
		ops->pgm_write_space0 = tpiWriteCS;
		ops->pgm_read_space1  = (uchar (*)(uint32_t))tpiReadDS;
		ops->pgm_write_space1 = tpiWriteDS;
#endif
	} else {
#ifdef ENABLE_PDI
		ops->pgm_poll         = pditpi_idle;
		ops->pgm_transmit     = pdiTransmit4;
		ops->pgm_enable       = pdiEnable;
		ops->pgm_read_space0  = (uchar (*)(uint16_t))pdiReadCS;
		ops->pgm_write_space0 = pdiWriteCS;
		ops->pgm_read_space1  = (uchar (*)(uint32_t))pdiReadDS;
		ops->pgm_write_space1 = pdiWriteDS;
		ops->pgm_blk_start    = pdiStartBlk;
		ops->pgm_open_page    = pdiOpenPage;
		ops->pgm_close_page   = pdiClosePage;
#endif
	}
#endif
}
